/* Copyright (C) 2003 Edson Pereira, PU1JTE, N1VTN, 7N4NCL
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, download a copy
 * from http://www.fsf.org
 *
 */

#include "petittrack.h"
#include "world0.xpm"
#include "world1.xpm"
#include "world2.xpm"

#ifndef SHARP_BUZZER_MAKESOUND
#define SHARP_BUZZER_MAKESOUND (0x5680)
#endif

PetitTrack::PetitTrack( QWidget* parent,  const char* name, WFlags fl )
   : QWidget( parent, name, fl )
{
   if ( !name ) setName( "PetitTrack" );
   resize( 196, 245 );
   setCaption( tr( "PetitTrack 1.2.0" ) );
   layout = new QVBoxLayout( this );
   layout->setSpacing( 1 );
   layout->setMargin( 0 );

   MyConfiguration = new Config(tr( "PetitTrack" ));
   MyConfiguration->setGroup( "petittrack group" );

   lat = atof( MyConfiguration->readEntry("lat", "0.0" ).latin1() );
   lon = atof( MyConfiguration->readEntry("lon", "0.0").latin1() );
   alt = atoi( MyConfiguration->readEntry("alt", "10").latin1() );
   mapType = atoi( MyConfiguration->readEntry( "map", "0").latin1() );
   rt1ObjectID = atoi( MyConfiguration->readEntry("obj", "0").latin1() );

   char *home = getenv( "HOME" );
   snprintf( tleFile, sizeof(tleFile), "%s/Documents/nasatle.txt", home );

   QFont font( "fixed", 12 );
   font.setFixedPitch( true );

   // Setup the Real Time View
   rtnView = new QListView( this, "textView" );
   rtnView->setFont( font );
   rtnView->addColumn( "Object", 68 );
   rtnView->addColumn( " ", 8 );
   rtnView->addColumn( "AOS (GMT)", 68 );
   rtnView->addColumn( "LOS (GMT)", 68 );
   rtnView->setSorting( 2 );
   rtnView->header()->setClickEnabled( false );
   rtnView->header()->setResizeEnabled( false );
   connect( rtnView, SIGNAL(pressed(QListViewItem *)),
	    this, SLOT(itemSelected( QListViewItem *)));

   // Setup the Single View
   rt1Frame = new QFrame( this );
   rt1ButtonGroup = new QButtonGroup( rt1Frame );
   rt1ButtonGroup->setExclusive( true );
   rt1ButtonGroup->setGeometry( 184, 0, 55, 264 );
   rt1ButtonGroup->setFrameShape( QFrame::NoFrame );
   connect( rt1ButtonGroup, SIGNAL(clicked(int)), this, SLOT(setRT1Object(int)) );

   QFrame *rt1LabelFrame = new QFrame( rt1Frame );
   rt1LabelFrame->setGeometry( 0, 121, 232, 133 );
   rt1LabelFrame->setFrameStyle( QFrame::Box | QFrame::Plain );

   QLabel *ll[14];

   for ( int i = 0; i < 10; i++ )
   {
      ll[i] = new QLabel( rt1LabelFrame );
      ll[i]->setFont( font );
      ll[i]->setGeometry( 1, 1 + ( 13 * i ), 70, 13 );

      rtLabel[i] = new QLabel( rt1LabelFrame );
      rtLabel[i]->setFont( font );
      rtLabel[i]->setGeometry( 70, 1+ ( 13 * i ), 161, 13 );
      if ( i % 2 == 0 ) rtLabel[i]->setBackgroundColor( QColor( 255,255,255 ) );
      else rtLabel[i]->setBackgroundColor( QColor( 220,220,255 ) );
   }

   aoslosLabel = ll[2];

   ll[0]->setText( "    Object: " );
   ll[1]->setText( " Date/Time: " );
   ll[2]->setText( "       AOS: " );
   ll[3]->setText( " CountDown: " );
   ll[4]->setText( "   Lat/Lon: " );
   ll[5]->setText( "     Az/El: " );
   ll[6]->setText( " Alt/Range: " );
   ll[7]->setText( " Phase/Rev: " );
   ll[8]->setText( "    UpLink: " );
   ll[9]->setText( "  DownLink: " );

   map0 = new QPixmap( (const char **)world0_xpm );
   map1 = new QPixmap( (const char **)world1_xpm );
   map2 = new QPixmap( (const char **)world2_xpm );
   worldPix = map0;
   
   QFrame *worldFrame = new QFrame( rt1Frame );
   worldFrame->setGeometry( 0, 2, worldPix->width() + 2, worldPix->height() + 2 );
   worldFrame->setFrameStyle( QFrame::Box | QFrame::Plain );

   worldMap = new QFrame( worldFrame );
   worldMap->setGeometry( 1, 1, worldPix->width(), worldPix->height() );

   initLabel = new QLabel( rt1Frame );
   initLabel->setGeometry( 0, 0, 232, 256 );
   initLabel->setAlignment( AlignHCenter );
   initLabel->setText( QString("\n\n\n\nPetitTrack v1.2.0\n\n\n") +
		       "Copyright (C) 2003 Edson Pereira\n" +
		       "\n\n\n\n\nInitializing...\n" );
   ephemFrame = new QFrame( this );

   QPushButton *ephemButton = new QPushButton( ephemFrame );
   ephemButton->setText( "Calculate" );
   ephemButton->setGeometry( 80, 237, 74, 20 );
   connect( ephemButton, SIGNAL(clicked()), this, SLOT(computeEphem()) );

   QPushButton *ephemStop = new QPushButton( ephemFrame );
   ephemStop->setText( "Stop" );
   ephemStop->setGeometry( 156, 237, 74, 20 );
   connect( ephemStop, SIGNAL(clicked()), this, SLOT(stopEphem()) );

   ephemStopFlag = false;
   
   QLabel *ephemDaysLabel = new QLabel( "Days", ephemFrame );
   ephemDaysLabel->setGeometry( 2, 237, 30, 20 );
   
   ephemDays = 7;
   ephemDaysSpin = new QSpinBox( ephemFrame, "ephemDaysSpin" );
   ephemDaysSpin->setGeometry( 33, 237, 45, 20 );
   ephemDaysSpin->setMinValue( 2 );
   ephemDaysSpin->setValue( ephemDays );
   connect( ephemDaysSpin, SIGNAL(valueChanged(int)), this, SLOT(setEphemDays(int)));
   
   ephemView = new QMultiLineEdit( ephemFrame );
   ephemView->setGeometry( 0, 2, 230, 235 );
   ephemView->setFont( font );
   ephemView->setReadOnly( true );

   QFrame *astroFrame = new QFrame( this );

   skyViewFrame = new QFrame( astroFrame );
   skyViewFrame->setGeometry( 2, 2, 230, 180 );
   skyViewFrame->setBackgroundColor( QColor( 0, 0, 100 ) );

   QFrame *skyFrame = new QFrame( astroFrame );
   skyFrame->setGeometry( 2, 182, 230, 77 );
   
   skyPix = new QPixmap( 230, 180 );

   QLabel *skySolLabel = new QLabel( "===== Sun ====", skyFrame );
   skySolLabel->setFont( font );
   skySolLabel->setGeometry( 43, 0, 90, 15 );
   
   QLabel *skyLunaLabel = new QLabel( "==== Moon ====", skyFrame );
   skyLunaLabel->setFont( font );
   skyLunaLabel->setGeometry( 139, 0, 90, 15 );
   
   QLabel *skyLabel2 = new QLabel( " Ele:", skyFrame );
   skyLabel2->setFont( font );
   skyLabel2->setGeometry( 4, 15, 43, 15 );

   skyEleLabel = new QLabel( skyFrame );
   skyEleLabel->setFont( font );
   skyEleLabel->setGeometry( 37, 15, 193, 15 );
   skyEleLabel->setBackgroundColor( QColor( 255, 255, 255 ) );
   
   QLabel *skyLabel1 = new QLabel( " Azi:", skyFrame );
   skyLabel1->setFont( font );
   skyLabel1->setGeometry( 4, 30, 43, 15 );
   
   skyAziLabel = new QLabel( skyFrame );
   skyAziLabel->setFont( font );
   skyAziLabel->setGeometry( 37, 30, 193, 15 );
   skyAziLabel->setBackgroundColor( QColor( 230, 255, 230 ) );
   
   QLabel *skyLabel3 = new QLabel( "Rise:", skyFrame );
   skyLabel3->setFont( font );
   skyLabel3->setGeometry( 4, 45, 43, 15 );
   
   skyLatLabel = new QLabel( skyFrame );
   skyLatLabel->setFont( font );
   skyLatLabel->setGeometry( 37, 45, 193, 15 );
   skyLatLabel->setBackgroundColor( QColor( 255, 255, 255 ) );
   
   QLabel *skyLabel4 = new QLabel( " Set:", skyFrame );
   skyLabel4->setFont( font );
   skyLabel4->setGeometry( 4, 60, 43, 15 );
   
   skyLonLabel = new QLabel( skyFrame );
   skyLonLabel->setFont( font );
   skyLonLabel->setGeometry( 37, 60, 193, 15 );
   skyLonLabel->setBackgroundColor( QColor( 230, 255, 230 ) );
   
   QFrame *configFrame = new QFrame( this );

   QGroupBox *coordinatesBox = new QGroupBox( "Coordinates", configFrame );
   coordinatesBox->setGeometry( 5, 2, 110, 80 );
      
   QLabel *lonLabel = new QLabel( "Lon", coordinatesBox );
   QLabel *latLabel = new QLabel( "Lat", coordinatesBox );
   QLabel *altLabel = new QLabel( "Alt", coordinatesBox );

   lonLine = new QLineEdit( coordinatesBox );
   latLine = new QLineEdit( coordinatesBox );
   altLine = new QLineEdit( coordinatesBox );

   lonLabel->setGeometry( 5, 15, 30, 17 );
   latLabel->setGeometry( 5, 35, 30, 17 );
   altLabel->setGeometry( 5, 55, 30, 17 );

   lonLine->setGeometry( 30, 15, 75, 17 );
   latLine->setGeometry( 30, 35, 75, 17 );
   altLine->setGeometry( 30, 55, 75, 17 );

   latLine->setText( MyConfiguration->readEntry("lat", "0.0" ).latin1() );
   lonLine->setText( MyConfiguration->readEntry("lon", "0.0" ).latin1() );
   altLine->setText( MyConfiguration->readEntry("alt", "0" ).latin1() );
   
   mapButtonGroup = new QButtonGroup( configFrame );
   mapButtonGroup->setTitle( "Map Type" );
   mapButtonGroup->setExclusive( true );
   
   QRadioButton *mapButton0 = new QRadioButton( "Full Color", mapButtonGroup );
   mapButton0->setGeometry( 5, 15, 100, 15 );
   QRadioButton *mapButton1 = new QRadioButton( "Two Color 1", mapButtonGroup );
   mapButton1->setGeometry( 5, 35, 100, 15 );
   QRadioButton *mapButton2 = new QRadioButton( "Two Color 2", mapButtonGroup );
   mapButton2->setGeometry( 5, 55, 100, 15 );

   mapButtonGroup->insert( mapButton0, 0 );
   mapButtonGroup->insert( mapButton1, 1 );
   mapButtonGroup->setGeometry( 120, 2, 110, 80 );
   connect( mapButtonGroup, SIGNAL(clicked(int)), this, SLOT(setMapType(int)));
   mapButtonGroup->setButton( mapType );
   
   QGroupBox *frequencyBox = new QGroupBox( "Radio Link (MHz)", configFrame );
   frequencyBox->setGeometry( 5, 85, 225, 60 );
      
   objTxRxLabel = new QLabel( "#####", frequencyBox );
   objTxRxLabel->setFrameStyle( QFrame::Panel | QFrame::Sunken );
   QLabel *txfLabel = new QLabel( " Up", frequencyBox );
   QLabel *rxfLabel = new QLabel( "Down", frequencyBox );

   txfLine = new QLineEdit( frequencyBox );
   rxfLine = new QLineEdit( frequencyBox );

   objTxRxLabel->setGeometry( 5, 15, 175, 17 );
   txfLabel->setGeometry( 5, 35, 30, 17 );
   rxfLabel->setGeometry( 115, 35, 30, 17 );
   txfLine->setGeometry( 30, 35, 75, 17 );
   rxfLine->setGeometry( 145, 35, 75, 17 );

   QPushButton *objTxRxNavL = new QPushButton( "<", frequencyBox );
   objTxRxNavL->setFont( font );
   objTxRxNavL->setGeometry( 185, 15, 15, 17 );
   connect( objTxRxNavL, SIGNAL(clicked()), this, SLOT(objTxRxNavigateLeft()) );

   QPushButton *objTxRxNavR = new QPushButton( ">", frequencyBox );
   objTxRxNavR->setFont( font );
   objTxRxNavR->setGeometry( 204, 15, 15, 17 );      
   connect( objTxRxNavR, SIGNAL(clicked()), this, SLOT(objTxRxNavigateRight()) );
   
   QFrame *aboutFrame = new QFrame( this );
   QLabel *aboutLabel = new QLabel( aboutFrame );
   aboutLabel->setGeometry( 20, 20, 200, 300 );
   aboutLabel->setText( QString( "PetitTrack v1.2.0\n\n" ) +
			"Copyright (c) 2003\n" +
			"Edson Pereira, PU1JTE, N1VTN, 7N4NCL\n\n" +
			"This program is free software; you can" + 
			"\nredistribute it and/or modify it under" +
			"\nthe terms of the GNU General Public " + 
			"\nLicense Version 2 as published by the" + 
			"\nFree Software Foundation." +
			"\n\nCredits: \n\n" +
			"Keplerian Propagator algorithm from\n" +
			"John A. Magliacane, KD2BD's Predict.\n\n\n\n\n\n" );

   tab = new QTabWidget( this, "tab" );
   tab->addTab( rt1Frame, "Tracker" );
   tab->addTab( astroFrame, "Sky" );
   tab->addTab( rtnView, "Queue" );
   tab->addTab( ephemFrame, "Ephemeris" );
   tab->addTab( configFrame, "Config" );
   tab->addTab( aboutFrame, "About" );
   connect( tab, SIGNAL(currentChanged(QWidget * )), this, SLOT(tabChanged(QWidget * )) );

   layout->addWidget( tab );

   for ( int i = 0; i < MAXOBJECTS; i++ )
   {
      aosTime[i] = 0;
      losTime[i] = 1;
   }

   // For the beep
   if ((beepfd = open("/dev/sharp_buz", O_WRONLY )) < 0)
   {
      perror("Unable to open /dev/sharp_buz");
   }
   
   // Initialize propagators
   loadTLE();
   
   objTxRxIndex = rt1ObjectID;   
   objTxRxLabel->setText( p[objTxRxIndex]->getSatName() );   

   char temp[20];
   
   sprintf( temp, "%f", p[objTxRxIndex]->getUplink() );
   txfLine->setText( temp );
   sprintf( temp, "%f", p[objTxRxIndex]->getDownlink() );
   rxfLine->setText( temp );
   
   groundTrackFlag = false;
   groundTrackPoints = new QPointArray( 60 );
   
   AzElPoints = new QPointArray( 60 );

   moon = new MoonPropagator( lat, lon );
   sun = new SunPropagator( lat, lon );   
   sunPoints = new QPointArray( 360 );

   sunRiseTime = calculateSunRiseTime( getDaynum() );
   sunSetTime = calculateSunSetTime( getDaynum() );
   moonRiseTime = calculateMoonRiseTime( getDaynum() );
   moonSetTime = calculateMoonSetTime( getDaynum() );
   
   QTimer *queueTimer = new QTimer( this );
   connect( queueTimer, SIGNAL(timeout()), this, SLOT(calculateQueue()));
   queueTimer->start( 1000, false );

   calculateSunFootprint();   
   QTimer *sunTimer = new QTimer( this );
   connect( sunTimer, SIGNAL(timeout()), this, SLOT(calculateSunFootprint()));
   sunTimer->start( 60000, false );

   propagate();
   t1 = new QTimer( this );
   connect( t1, SIGNAL(timeout()), this, SLOT(propagate()));
   t1->start( 1000, false );

   calculateGroundTrack();
   QTimer *groundtrackTimer = new QTimer( this );
   connect( groundtrackTimer, SIGNAL(timeout()), this, SLOT(calculateGroundTrack()));
   groundtrackTimer->start( 120000, false );
   
   setMapType( mapType );
   drawMap();
   QTimer *mapTimer = new QTimer( this );
   connect( mapTimer, SIGNAL(timeout()), this, SLOT(drawMap()));
   mapTimer->start( 10000, false );

   ephemTimer = new QTimer( this );
   connect( ephemTimer, SIGNAL(timeout()), this, SLOT(runEphem()));
   calculating = false;   
}

PetitTrack::~PetitTrack()
{
   char tmp[64];
   char key[64];
   bool ok;

   sprintf( tmp, "%d", rt1ObjectID );
   MyConfiguration->writeEntry( "obj", tmp );

   sprintf( tmp, "%d", mapType );
   MyConfiguration->writeEntry( "map", tmp );

   latLine->text().toDouble( &ok );
   if ( ok ) MyConfiguration->writeEntry( "lat", latLine->text() );

   lonLine->text().toDouble( &ok );
   if ( ok ) MyConfiguration->writeEntry( "lon", lonLine->text() );

   altLine->text().toInt( &ok );
   if ( ok ) MyConfiguration->writeEntry( "alt", altLine->text() );


   for ( int i = 0; i < numObjects; i++ )
   {
      sprintf( tmp, "%f", p[i]->getUplink() );
      sprintf( key, "tx_%ld", p[i]->getCatalogNumber() );	 
      MyConfiguration->writeEntry( key, tmp );
      sprintf( tmp, "%f", p[i]->getDownlink() );
      sprintf( key, "rx_%ld", p[i]->getCatalogNumber() );	 
      MyConfiguration->writeEntry( key, tmp );      
   }
   
   delete MyConfiguration;
}

void PetitTrack::loadTLE()
{
   FILE *fp;
   char line0[ 81 ];
   char line1[ 81 ];
   char line2[ 81 ];
   char tmpname[ 81 ];
   char fragname[ 81 ];
   char index[ 3 ];
   int linenum = 0;
   unsigned int i;
   char *r;
   char temp[256];
   double f;
   
   numObjects = 0;

   fp = fopen( tleFile, "r" );
   if ( fp == NULL )
   {
      perror( tleFile );
      QMessageBox::critical( 0, "PetitTrack",
			     QString("PetitTrack was unable to find the file\n") + tleFile +
			     "\n\nPlease download a recent" +
			     "\nNASA TLE file from " +
			     "\nhttp://www.celestrak.com or" +
			     "\nany other source and place" +
			     "\nthe file in the directory" + 
			     "\n" + getenv("HOME") + "/Documents." +
			     "\n\nThe file must be named" +
			     "\nnasatle.txt" +
			     "\n\nPetitTrack will now exit." );
      exit(1);
   }

   while( numObjects < MAXOBJECTS )
   {
      r = fgets( line0, 80, fp );
      if ( r == NULL ) break;
      if ( line0[0] == '\n' ) continue;
      linenum++;

      // Clean up the trailing spaces 
      for ( i = 0; i < strlen(line0); i++ )
      {
	 if ( line0[i] == '\r' || line0[i] == '\n' ) 
	    line0[i] = '\0';
      }
      for ( i = strlen( line0 ) - 1; i > 0; i-- )
      {
	 if ( line0[i] != ' ' ) 
	 {
	    line0[i+1] = '\0';
	    break;
	 }
      }
      
      // If name is longer than 11 characters, shrink it
      if ( strlen( line0 ) > 11 )
      {
	 strcpy( fragname, line0 );
	 fragname[5] = '\0';	    
	 sprintf( tmpname, "%s~%s", fragname, line0 + strlen(line0) - 5 );
	 strcpy( line0, tmpname );
      }
      
      fgets( line1, 80, fp );
      if ( line1[0] != '1' )
      {
	 sprintf( temp, "Unexpected error on TLE file line %d\n", linenum );
	 QMessageBox::critical( 0, "PetitTrack",
				QString( temp ) + "\n\n" +
				"TLE File: \n" + tleFile +
				"\n\nPlease check the TLE File integrity" +
				"\n\nPetitTrack will now exit." );
	 exit(1);
      }
      linenum++;

      fgets( line2, 80, fp );
      if ( line2[0] != '2' )
      {
	 sprintf( temp, "Unexpected error on TLE file line %d\n", linenum );
	 QMessageBox::critical( 0, "PetitTrack",
				QString( temp ) + "\n\n" +
				"TLE File: \n" + tleFile +
				"\n\nPlease check the TLE File integrity" +
				"\n\nPetitTrack will now exit." );
	 exit(1);
      }
      linenum++;

      p[ numObjects ] = new KepPropagator();
      p[ numObjects ]->init( line0, line1, line2, lat, lon, alt );

      sprintf( temp, "tx_%ld", p[ numObjects ]->getCatalogNumber() );
      f = atof( MyConfiguration->readEntry( temp, "145.9" ).latin1() );
      p[ numObjects ]->setUplink( f );
      
      sprintf( temp, "rx_%ld", p[ numObjects ]->getCatalogNumber() );
      f = atof( MyConfiguration->readEntry( temp, "435.5" ).latin1() );
      p[ numObjects ]->setDownlink( f );

      sprintf( index, "%d", numObjects );
      listViewItem[ numObjects ] = new QListViewItem( rtnView, line0, " ", "?", "?", QString( index ) );

      numObjects++;
   }
   
   if ( rt1ObjectID > numObjects ) rt1ObjectID = 0;
   rtnView->setCurrentItem( listViewItem[ rt1ObjectID ] );
}

void PetitTrack::findAOSLOS( double daynum, KepPropagator *kp, int satIndex )
{
   bool inview = false;
   double initialDaynum;
   bool found;
   
   if ( kp->Geostationary() )
   {
      aosTime[satIndex] = -1.0;
      losTime[satIndex] = -1.0;
      kp->propagate( daynum );
      if ( kp->getElevation() > 0.0 )
      {
	 listViewItem[satIndex]->setText( 2, "Always" );
	 listViewItem[satIndex]->setText( 3, " " );
      }
      else
      {
	 listViewItem[satIndex]->setText( 2, "Never" );
	 listViewItem[satIndex]->setText( 3, " " );
      }
      return;
   }
   
   if ( aosTime[satIndex] > -2.0 && !kp->AosHappens() )
   {
      aosTime[satIndex] = -3.0;
      losTime[satIndex] = -3.0;
      listViewItem[satIndex]->setText( 2, "No AOS" );
      listViewItem[satIndex]->setText( 3, "No LOS " );
      return;
   }
   
   for ( initialDaynum = daynum, found = false; 
	 found == false;
	 daynum = incrementDaynum( daynum, kp ) )
   {
      kp->propagate( daynum );

      if ( !inview && kp->getElevation() > 0.0 )
      {
	 inview = true;
	 aosTime[ satIndex ] = daynum;
      }
      
      if ( inview && kp->getElevation() < 0.0 )
      {
	 inview = false;
	 losTime[ satIndex ] = daynum;
	 found = true;
      }
      
      if ( (daynum - initialDaynum) > 7.0 )
      {
	 aosTime[satIndex] = -5.0;
	 losTime[satIndex] = -5.0;	 
	 listViewItem[satIndex]->setText( 2, "> 7 Days" );
	 listViewItem[satIndex]->setText( 3, " " );
	 found = true;
      }
   }
}

void PetitTrack::calculateQueue()
{
   double daynum;   
   int i;
   char temp[256];
   struct tm *aos_tm;
   struct tm *los_tm;
   time_t t;
   bool flag;
   static bool firstTime = true;
   
   flag = false;
   
   daynum = getDaynum();
   
   for ( i = (firstTime == true) ? rt1ObjectID : 0; i < numObjects; i++ )
   {
      if ( (losTime[i] < daynum) && (losTime[i] > 0.0) )
      {
	 findAOSLOS( daynum, p[i], i );
	 
	 printf("%2d %f %s\n", i, losTime[i], p[i]->getSatName() );
	 
	 if ( aosTime[i] > 0.0 )
	 {
	    t = p[i]->daynum2unix( aosTime[i] );
	    aos_tm = gmtime( &t );
	    strftime( temp, 20, "%m%d %R",  aos_tm );
	    listViewItem[i]->setText( 2, temp );
	    
	    t = p[i]->daynum2unix( losTime[i] );
	    los_tm = gmtime( &t );
	    strftime( temp, 20, "%m%d %R",  los_tm );
	    listViewItem[i]->setText( 3, temp );

	    rtnView->sort();
	 }
	 
	 flag = true;
      }
      
      if ( daynum > aosTime[i] && aosTime[i] > 0.0 ) 
	 sprintf( temp, "*" );
      else 
	 sprintf( temp, " " );
      
      listViewItem[i]->setText( 1, temp );
      
      // Do not process more than one object per timer tick.
      // This is mainly for graceful startup when all objects
      // need to have their AOS/LOS recalculated.
      if ( flag ) return;

      if ( firstTime ) 
      {
	 firstTime = false;
	 initLabel->hide();	 
      }      
   }   
}

double PetitTrack::getDaynum()
{
   struct timeb tptr;

   ftime( &tptr );
   return ((((double)tptr.time + 0.001 * (double)tptr.millitm) / 86400.0 ) - 3651.0);
}

void PetitTrack::propagate()
{
   double daynum;
   char temp[256];
   char temp3[4];
   struct tm *aos_tm;
   struct tm *los_tm;
   struct tm *temp_tm;
   time_t t;
   static int blink = 0;
   
   daynum = getDaynum();

   p[rt1ObjectID]->propagate( daynum );

   if ( p[rt1ObjectID]->getElevation() >= 0.0 )
   {
      sprintf( temp, "  %-11s [ VISIBLE ]", p[rt1ObjectID]->getSatName() );
      rtLabel[0]->setBackgroundColor( QColor( 220,255,220 ) );
   }
   else
   {
      sprintf( temp, "  %-9s", p[rt1ObjectID]->getSatName() );
      rtLabel[0]->setBackgroundColor( QColor( 255,255,255 ) );
   }
   rtLabel[0]->setText( temp );

   t = p[rt1ObjectID]->daynum2unix( daynum );
   temp_tm = gmtime( &t );
   strftime( temp, 40, "  %Y%m%d  %T GMT",  temp_tm );
   rtLabel[1]->setText( temp );

   if ( p[rt1ObjectID]->getElevation() > 0.0 )
   {
      if ( losTime[rt1ObjectID] < 0.0 )
      {
	 if ( losTime[rt1ObjectID] < -4.0 )
	    sprintf( temp, "  > 7 Days");	 
	 else if ( losTime[rt1ObjectID] < 2.0 )
	    sprintf( temp, "  Never");
	 
	 if ( p[rt1ObjectID]->Geostationary() )
	    sprintf( temp, "  Geostationary Orbit");	 
      }
      else
      {
	 t = p[rt1ObjectID]->daynum2unix( losTime[rt1ObjectID] );
	 los_tm = gmtime( &t );
	 strftime( temp, 40, "  %Y%m%d  %T GMT",  los_tm );
      }
      rtLabel[2]->setText( temp );
      sprintf( temp, "       LOS: " );
      
      // Blink for one minute
      if ( fabs( losTime[rt1ObjectID] - daynum ) < 0.000694 )
      {
	 ioctl( beepfd, SHARP_BUZZER_MAKESOUND, 2 );
	 if (( blink++ % 2) == 0 ) 
	    sprintf( temp3, "LOS" );
	 else
	    sprintf( temp3, "   " );
      }
      else 
      {
	 if ( blink != 0 )
	    ioctl( beepfd, SHARP_BUZZER_MAKESOUND, 11 );
	 
	 blink = 0;
	 sprintf( temp3, "LOS" );
      }
      
      aoslosLabel->setText( temp );
   }
   else
   {
      if ( aosTime[rt1ObjectID] < 0.0 )
      {
	 if ( losTime[rt1ObjectID] < -4.0 )
	    sprintf( temp, "  > 7 Days");	 
	 else if ( losTime[rt1ObjectID] < 2.0 )
	    sprintf( temp, "  Never");
	 
	 if ( p[rt1ObjectID]->Geostationary() )
	    sprintf( temp, "  Geostationary Orbit");	 
      }
      else
      {
	 t = p[rt1ObjectID]->daynum2unix( aosTime[rt1ObjectID] );
	 aos_tm = gmtime( &t );
	 strftime( temp, 40, "  %Y%m%d  %T GMT",  aos_tm );
      }
      rtLabel[2]->setText( temp );
      sprintf( temp, "       AOS: " );

      // Blink for one minute
      if ( fabs( aosTime[rt1ObjectID] - daynum ) < 0.000694 )
      {
	 ioctl( beepfd, SHARP_BUZZER_MAKESOUND, 2 );

	 if (( blink++ % 2) == 0 ) 
	    sprintf( temp3, "AOS" );
	 else
	    sprintf( temp3, "   " );
      }
      else 
      {
	 if ( blink != 0 )
	    ioctl( beepfd, SHARP_BUZZER_MAKESOUND, 11 );
	 
	 blink = 0;
	 sprintf( temp3, "AOS" );
      }
      
      aoslosLabel->setText( temp );
   }
  
   // The labels in the Sky View
   sun->propagate( daynum );
   moon->propagate( daynum );
   
   sprintf( temp, "%7.2f%c        %7.2f%c", sun->getEl(), 0xb0, moon->getEl(), 0xb0 );
   skyEleLabel->setText( temp );
   
   sprintf( temp, "%7.2f%c        %7.2f%c", sun->getAz(), 0xb0, moon->getAz(), 0xb0 );
   skyAziLabel->setText( temp );
   
   char tempS[20];
   char tempM[20];
   
   if ( sun->getEl() < 0.0 && sunRiseTime < daynum )
      sunRiseTime = calculateSunRiseTime( daynum );
   
   if ( moon->getEl() < 0.0 && moonRiseTime < daynum )
      moonRiseTime = calculateMoonRiseTime( daynum );
      
   t = p[rt1ObjectID]->daynum2unix( sunRiseTime );
   temp_tm = localtime( &t );
   strftime( tempS, 40, " %m%d %H:%M %Z", temp_tm );
   t = p[rt1ObjectID]->daynum2unix( moonRiseTime );
   temp_tm = localtime( &t );
   strftime( tempM, 40, " %m%d %H:%M %Z", temp_tm );
   sprintf( temp, "%s %s", tempS, tempM );
   skyLatLabel->setText( temp );
   
   if ( sun->getEl() > 0.0 && sunSetTime < daynum )
      sunSetTime = calculateSunSetTime( daynum );
   
   if ( moon->getEl() > 0.0 && moonSetTime < daynum )
      moonSetTime = calculateMoonSetTime( daynum );
      
   t = p[rt1ObjectID]->daynum2unix( sunSetTime );
   temp_tm = localtime( &t );
   strftime( tempS, 40, " %m%d %H:%M %Z", temp_tm );
   t = p[rt1ObjectID]->daynum2unix( moonSetTime );
   temp_tm = localtime( &t );
   strftime( tempM, 40, " %m%d %H:%M %Z", temp_tm );
   sprintf( temp, "%s %s", tempS, tempM );
   skyLonLabel->setText( temp );

   // The labels in the map view
   double countDownTime1;
   
   if ( aosTime[rt1ObjectID] > 0.0 )
   {   
      if ( p[rt1ObjectID]->getElevation() < -0.2 ) 
	 countDownTime1 = aosTime[rt1ObjectID] - daynum + 1.0;
      else 
	 countDownTime1 = losTime[rt1ObjectID] - daynum + 1.0;

      t = p[rt1ObjectID]->daynum2unix( countDownTime1 );

      sprintf( countDown, " " );	 

      if ( t > (86400*365 + 315532800) )
      {
	 sprintf( temp, " >7 days" );
      }
      else
      {
	 temp_tm = gmtime( &t );
	 sprintf( temp, "  %3d day%s  %02d:%02d:%02d",
		  temp_tm->tm_yday,
		  temp_tm->tm_yday == 1 ? " " : "s", 
		  temp_tm->tm_hour,
		  temp_tm->tm_min,
		  temp_tm->tm_sec );

	 // Save for the Sky view
	 sprintf( countDown, "%3d %02d:%02d:%02d %s",
		  temp_tm->tm_yday,
		  temp_tm->tm_hour,
		  temp_tm->tm_min,
		  temp_tm->tm_sec,
		  temp3);
      }
   }
   else
   {
      sprintf( temp, "         0  00:00:00" );
   }
    
   rtLabel[3]->setText( temp );

   sprintf( temp, " %5.1f%c %c   %5.1f%c %c",
	    p[rt1ObjectID]->getSspLat() > 0.0 ? p[rt1ObjectID]->getSspLat() : p[rt1ObjectID]->getSspLat() * -1.0, 
	    0xb0, p[rt1ObjectID]->getSspLat() > 0.0 ? 'N' : 'S',
	    p[rt1ObjectID]->getSspLon() > 180.0 ? 360.0 - p[rt1ObjectID]->getSspLon() : p[rt1ObjectID]->getSspLon(), 
	    0xb0, p[rt1ObjectID]->getSspLon() > 180.0 ? 'E' : 'W' );
   rtLabel[4]->setText( temp );

   sprintf( temp, " %5.1f%c     %5.1f%c",
	    p[rt1ObjectID]->getAzimuth(), 0xb0,
	    p[rt1ObjectID]->getElevation(), 0xb0);
   rtLabel[5]->setText( temp );

   sprintf( temp, " %5.0f km   %5.0f km",
	    p[rt1ObjectID]->getAltitude(),
	    p[rt1ObjectID]->getSlantRange());
   rtLabel[6]->setText( temp );
   
   sprintf( temp, "   %3.0f     %6ld",
	    p[rt1ObjectID]->getPhase() * 256,
	    p[rt1ObjectID]->getOrbitNumber());
   rtLabel[7]->setText( temp );

   sprintf( temp, " %12.6f MHz %3.1f dB", 
	    p[rt1ObjectID]->getUplink() - p[rt1ObjectID]->getDoppler( p[rt1ObjectID]->getUplink() ),
	    p[rt1ObjectID]->getPathLoss( p[rt1ObjectID]->getUplink() ));
   rtLabel[8]->setText( temp );

   sprintf( temp, " %12.6f MHz %3.1f dB",  
	    p[rt1ObjectID]->getDownlink() + p[rt1ObjectID]->getDoppler( p[rt1ObjectID]->getDownlink() ),
	    p[rt1ObjectID]->getPathLoss( p[rt1ObjectID]->getDownlink() ));
   rtLabel[9]->setText( temp );
   
   drawSky();
}

double PetitTrack::calculateSunRiseTime( double daynum )
{   
   SunPropagator sp( lat, lon );
   double step;
   
   sp.propagate( daynum );
   if ( sp.getEl() > 0.0 )
   {
      do
      {
	 if ( sp.getEl() > 5.0 )
	    step = 0.02083;  // 30 minutes 
	 else if ( sp.getEl() > 0.2 )
	    step = 0.000694; // 1 minute
	 else
	    step = 0.000012; // 1 second
	 daynum -= step;
	 //printf( "^ SR %f %.10f %.10f\n", daynum, sp.getEl(), step );
	 sp.propagate( daynum );
      }
      while ( sp.getEl() > 0.0 );
   }
   else
   {
      do
      {
	 if ( sp.getEl() < -5.0 )
	    step = 0.02083;  // 30 minutes 
	 else if ( sp.getEl() < -0.2 )
	    step = 0.000694; // 1 minute
	 else
	    step = 0.000012; // 1 second
	 daynum += step;
	 //printf( "v SR %f %.10f %.10f\n", daynum, sp.getEl(), step );
	 sp.propagate( daynum );
      }
      while ( sp.getEl() < 0.0 );
   }
   //printf("Done -----------------------\n");
   return daynum;
}

double PetitTrack::calculateSunSetTime( double daynum )
{
   SunPropagator sp( lat, lon );
   double step;
   
   sp.propagate( daynum );
   if ( sp.getEl() > 0.0 )
   {
      do
      {
	 if ( sp.getEl() > 5.0 )
	    step = 0.02083;  // 30 minutes 
	 else if ( sp.getEl() > 0.2 )
	    step = 0.000694; // 1 minute
	 else
	    step = 0.000012; // 1 second
	 daynum += step;
	 //printf( "^ SS %f %.10f %.10f\n", daynum, sp.getEl(), step );
	 sp.propagate( daynum );
      }
      while ( sp.getEl() > 0.0 );
   }
   else
   {
      do
      {
	 if ( sp.getEl() < -5.0 )
	    step = 0.02083;  // 30 minutes 
	 else if ( sp.getEl() < -0.2 )
	    step = 0.000694; // 1 minute
	 else
	    step = 0.000012; // 1 second
	 daynum -= step;
	 //printf( "v SS %f %.10f %.10f\n", daynum, sp.getEl(), step );
	 sp.propagate( daynum );
      }
      while ( sp.getEl() < 0.0 );
   }
   //printf("Done -----------------------\n");
   return daynum;   
}

double PetitTrack::calculateMoonRiseTime( double daynum )
{
   MoonPropagator mp( lat, lon );
   double step;
   
   mp.propagate( daynum );
   if ( mp.getEl() > 0.0 )
   {
      do
      {
	 if ( mp.getEl() > 5.0 )
	    step = 0.02083;  // 30 minutes 
	 else if ( mp.getEl() > 0.2 )
	    step = 0.000694; // 1 minute
	 else
	    step = 0.000012; // 1 second
	 daynum -= step;
	 //printf( "^ MR %f %.10f %.10f\n", daynum, mp.getEl(), step );
	 mp.propagate( daynum );
      }
      while ( mp.getEl() > 0.0 );
   }
   else
   {
      do
      {
	 if ( mp.getEl() < -5.0 )
	    step = 0.02083;  // 30 minutes 
	 else if ( mp.getEl() < -0.2 )
	    step = 0.000694; // 1 minute
	 else
	    step = 0.000012; // 1 second
	 daynum += step;
	 //printf( "v MR %f %.10f %.10f\n", daynum, mp.getEl(), step );
	 mp.propagate( daynum );
      }
      while ( mp.getEl() < 0.0 );
   }
   //printf("Done -----------------------\n");
   return daynum;   
}

double PetitTrack::calculateMoonSetTime( double daynum )
{
   MoonPropagator mp( lat, lon );
   double step;
   
   mp.propagate( daynum );
   if ( mp.getEl() > 0.0 )
   {
      do
      {
	 if ( mp.getEl() > 5.0 )
	    step = 0.02083;  // 30 minutes 
	 else if ( mp.getEl() > 0.2 )
	    step = 0.000694; // 1 minute
	 else
	    step = 0.000012; // 1 second
	 daynum += step;
	 //printf( "^ MS %f %.10f %.10f\n", daynum, mp.getEl(), step );
	 mp.propagate( daynum );
      }
      while ( mp.getEl() > 0.0 );
   }
   else
   {
      do
      {
	 if ( mp.getEl() < -5.0 )
	    step = 0.02083;  // 30 minutes 
	 else if ( mp.getEl() < -0.2 )
	    step = 0.000694; // 1 minute
	 else
	    step = 0.000012; // 1 second
	 daynum -= step;
	 //printf( "v MS %f %.10f %.10f\n", daynum, mp.getEl(), step );
	 mp.propagate( daynum );
      }
      while ( mp.getEl() < 0.0 );
   }
   //printf("Done -----------------------\n");
   return daynum;      
}

void PetitTrack::drawMap()
{
   KepPropagator *kp;
   QPen pen1, pen2, pen3, pen4, pen5, pen6;
   QPointArray pa( 360 );
   int x, y, xx, yy;

   // Fragments of the code below derived from KD2DB's Predict
   double spiderlat, spiderlong, azimuth, ssplat, ssplong, TWOPI, R0=6378.16,
      beta, num, dem;

   int azi;

   float height = (float)worldPix->height();
   float width = (float)worldPix->width();

   kp = p[rt1ObjectID];

   TWOPI = 2.0 * M_PI;

   ssplat = kp->getSspLat() * deg2rad;
   ssplong = kp->getSspLon() * deg2rad;
   beta = 0.5 * kp->getFootprint() / R0;

   /* Compute the visibility circle */
   for ( azi = 0; azi < 360; azi++)
   {
      azimuth = deg2rad * (double)azi;
      spiderlat = asin( sin( ssplat ) * cos( beta )
			+ cos( azimuth ) * sin( beta ) * cos( ssplat ));
      num = cos( beta ) - ( sin( ssplat ) * sin( spiderlat ));
      dem = cos( ssplat ) * cos( spiderlat );

      if ( azi == 0 && ( beta > M_PI_2 - ssplat ))
	 spiderlong = ssplong + M_PI;
      else if ( azi == 180 && ( beta > M_PI_2 + ssplat ))
	 spiderlong = ssplong + M_PI;
      else if ( fabs( num / dem) > 1.0 )
	 spiderlong = ssplong;
      else
      {
	 if (( 180 - azi ) >= 0)
	    spiderlong = ssplong - arccos( num, dem );
	 else
	    spiderlong = ssplong + arccos( num, dem );
      }

      while ( spiderlong < 0.0 )
	 spiderlong += TWOPI;

      while ( spiderlong > ( 2.0 * M_PI ))
	 spiderlong -= TWOPI;

      spiderlat = spiderlat / deg2rad;
      spiderlong = spiderlong / deg2rad;

      /* Convert spiderweb data to map-based coordinates */
      if ( spiderlong <= 180.0 )
	 xx = (int)( height - ( width * ( spiderlong / 360.0 )));
      else
      {
	 spiderlong = spiderlong - 180.0;
	 xx = (int)( width - ( width * ( spiderlong / 360.0 )));
      }

      yy = (int)( height / 2.0 - ( height / 2.0 * ( spiderlat / 90.0 )));
      pa.setPoint( azi, xx, yy );
   }

   // The sub satellite point map coordinates 
   if ( kp->getSspLon() <= 180.0) x = (int)( height - (width * ( kp->getSspLon() / 360.0 )));
   else x = (int)( width - ( width * ( (kp->getSspLon() - 180.0) / 360.0 )));

   y = (int)( height / 2.0 - ( height / 2.0 * ( kp->getSspLat() / 90.0 )));

   // The observer map coordinates
   int lonx, laty;
   if ( lon < 0.0 ) lonx = (int)( width - ( width * (( 180.0 - fabs(lon)) / 360.0 )));
   else lonx = (int)(  ( width * (( 180.0 - lon )/ 360.0 )));

   laty = (int)( height / 2.0 - ( height / 2.0 * ( lat / 90.0 )));

   QPainter p;
   QPixmap tmpMap = *worldPix;
   p.begin( &tmpMap );

   // Sharpen the pencils
   pen1.setWidth( 1 );
   pen2.setWidth( 1 );
   pen3.setWidth( 1 );
   pen4.setWidth( 1 );
   pen5.setWidth( 1 );
   pen6.setWidth( 1 );

   // Kindergarten time!
   pen1.setColor( QColor( 255, 0, 0 ) );
   pen2.setColor( QColor( 0, 200, 255 ) );
   pen3.setColor( QColor( 255, 255, 0 ) );
   pen4.setColor( QColor( 0, 255, 0 ) );
   pen5.setColor( QColor( 255, 255, 255 ) );
   pen6.setColor( QColor( 150, 150, 150 ) );
   
   // Draw the satellite trajectory
   p.setPen( pen4 );
   p.drawPoints( *groundTrackPoints );

   // Draw the sun terminator
   p.setPen( pen6 );
   p.drawPoints( *sunPoints );

   // Draw the observer
   p.setPen( pen2 );
   p.drawRect( lonx - 1, laty - 1, 3, 3 );
   
   // Draw the satellite
   p.setPen( pen1 );
   p.drawRect( x - 1, y - 1, 3, 3 );

   // Draw the satellite footprint
   p.drawPoints( pa );

   // Draw the <moon
   p.setPen( pen5 );
   p.drawRect( moon_x - 1, moon_y - 1, 3, 3 );
   
   // Draw the sun 
   p.setPen( pen3 );
   p.drawRect( sun_x - 1, sun_y - 1, 3, 3 );
   
   p.end();
   
   // Move map offscreen buffer to screen
   worldMap->setBackgroundPixmap( tmpMap );
}

void PetitTrack::drawSky()
{
   double azimuth;
   KepPropagator *kp;
   QPen pen1, pen2, pen3, pen4, pen5, pen6;
   QPainter pSky;
   QPixmap tmpSky( 230, 180 );

   kp = p[rt1ObjectID];
   
   if ( sun->getEl() > 0 )
      tmpSky.fill( QColor( 0, 0, 100 ) );
   else
      tmpSky.fill( QColor( 0, 0, 0 ) );
      
   // Sharpen the pencils
   pen1.setWidth( 1 );
   pen2.setWidth( 1 );
   pen3.setWidth( 1 );
   pen4.setWidth( 1 );
   pen5.setWidth( 1 );
   pen6.setWidth( 1 );

   // Kindergarten time!
   pen1.setColor( QColor( 255, 0, 0 ) );
   pen2.setColor( QColor( 0, 200, 255 ) );
   pen3.setColor( QColor( 255, 255, 0 ) );
   pen4.setColor( QColor( 0, 255, 0 ) );
   pen5.setColor( QColor( 255, 255, 255 ) );
   pen6.setColor( QColor( 150, 150, 150 ) );
   
   pSky.begin( &tmpSky );
   
   // Draw the circular scale
   pSky.setPen( pen6 );
   pSky.drawArc( 40, 15, 149, 149, 360*16, 360*16 );
   pSky.drawArc( 65, 40, 99, 99, 360*16, 360*16 );
   pSky.drawArc( 90, 65, 49, 49, 360*16, 360*16 );
   pSky.drawPoint( 115, 90 );

   // Draw the cardinal legends
   pSky.setFont( QFont( "fixed", 12 ) );
   pSky.drawText( 112, 13, "N" );
   pSky.drawText( 112, 175, "S" );
   pSky.drawText( 32, 94, "E" );
   pSky.drawText( 193, 94, "W" );
   
   // Draw the orbit path
   pSky.setPen( pen4 );
   pSky.drawPoints( *AzElPoints );
   
   int skyx, skyy;
   
   // Draw the sun
   if ( sun->getEl() > 0 )
   {
      azimuth=(double)( sun->getAz() - 90.0) * deg2rad;
      skyx=(int)((149/2.0) - cos(azimuth)*((double)(149/2.0)-sun->getEl()*(double)(149/2.0)/90.0));
      skyy=(int)((149/2.0) + sin(azimuth)*((double)(149/2.0)-sun->getEl()*(double)(149/2.0)/90.0));
      pSky.setPen( pen3 );      
      pSky.drawRect( 40 + skyx - 1, 15 + skyy - 1, 3, 3 );
   }

   // Draw the moon
   if ( moon->getEl() > 0 )
   {
      azimuth=(double)( moon->getAz() - 90.0) * deg2rad;
      skyx=(int)((149/2.0) - cos(azimuth)*((double)(149/2.0)-moon->getEl()*(double)(149/2.0)/90.0));
      skyy=(int)((149/2.0) + sin(azimuth)*((double)(149/2.0)-moon->getEl()*(double)(149/2.0)/90.0));
      pSky.setPen( pen5 );      
      pSky.drawRect( 40 + skyx - 1, 15 + skyy - 1, 3, 3 );
   }
   
   // Draw the satellite
   if ( kp->getElevation() > 0 )
   {
      azimuth=(double)( kp->getAzimuth() - 90.0) * deg2rad;
      skyx=(int)((149/2.0) - cos(azimuth)*((double)(149/2.0)-kp->getElevation()*(double)(149/2.0)/90.0));
      skyy=(int)((149/2.0) + sin(azimuth)*((double)(149/2.0)-kp->getElevation()*(double)(149/2.0)/90.0));
      pSky.setPen( pen1 );      
      pSky.drawRect( 40 + skyx - 1, 15 + skyy - 1, 3, 3 );
   }
   
   struct tm *temp_tm;
   time_t t;
   char temp[40];
   
   t = time( NULL );
   
   // Draw the UTC time 
   temp_tm = localtime( &t );
   strftime( temp, 40, "%T %Z", temp_tm );
   pSky.setPen( pen2 );      
   pSky.drawText( 154, 175, temp );
   
   // Draw the countdown
   if ( aosTime[rt1ObjectID] > 0.0 || kp->getElevation() > 0.0 )
   {
      pSky.setPen( pen4 );      
      pSky.drawText( 130, 13, countDown );
   
      // Draw the satellite legend
      pSky.setPen( pen1 ); 
      pSky.drawRect( 5, 7, 3, 3 );
      sprintf( temp, "%s", kp->getSatName() );
      pSky.setPen( pen5 );
      pSky.drawText( 13, 13, temp );
   
      // Draw the satellite range legend
      sprintf( temp, "R %.0f km", kp->getSlantRange() );
      pSky.setPen( pen5 );
      pSky.drawText( 5, 28, temp );
      
      // Draw the satellite azimuth legend
      sprintf( temp, "AZ %.0f%c", kp->getAzimuth(), 0xb0 );
      pSky.setPen( pen5 );
      pSky.drawText( 5, 43, temp );
   
      // Draw the satellite elevation legend
      sprintf( temp, "EL %.0f%c", kp->getElevation(), 0xb0 );
      pSky.setPen( pen5 );
      pSky.drawText( 5, 58, temp );
   }
   
   // Draw the sun legend
   pSky.setPen( pen3 );
   pSky.drawRect( 5, 154, 3, 3 );
   sprintf( temp, "Sun" );
   pSky.setPen( pen5 );
   pSky.drawText( 13, 160, temp );

   // Draw the moon legend
   pSky.setPen( pen5 );      
   pSky.drawRect( 5, 169, 3, 3 );
   sprintf( temp, "Moon" );
   pSky.setPen( pen5 );
   pSky.drawText( 13, 175, temp );
   
   pSky.end();

   // Display the sky view
   skyViewFrame->setBackgroundPixmap( tmpSky );
}

double PetitTrack::arccos( double x, double y )
{

   /* This function implements the arccosine function,
      returning a value between 0 and two pi.
      The code is borrowed from Predict.
    */

   double result = 0, fraction;

   fraction=x/y;

   if ((x>0.0) && (y>0.0))
      result=acos(fraction);

   if ((x<0.0) && (y>0.0))
      result=acos(fraction);

   if ((x<0.0) && (y<0.0))
      result = 3.14159 +  acos(fraction);

   if ((x>0.0) && (y<0.0))
      result = 3.14159 + acos(fraction);

   return result;
}

void PetitTrack::setRT1Object( int id )
{
   rt1ObjectID = id;
}

double PetitTrack::incrementDaynum( double daynum, KepPropagator *kp )
{
   double delta;
   static int count = 0;
   
   delta = 0;

   if ( kp->getEccentricity() > 0.3 )
   {
      if ( fabs( kp->getElevation() ) < 0.5 ) 
	 delta = 0.000012 / kp->getMeanMotion() *  kp->getAltitude() / 3000.0;
      else if ( fabs( kp->getElevation() ) > 0.5 && fabs( kp->getElevation() ) < 5.0 ) 
	 delta = 0.0005 / kp->getMeanMotion() *  kp->getAltitude() / 3000.0;
      else if ( fabs( kp->getElevation() ) > 5.0 && fabs( kp->getElevation() ) < 15.0 ) 
	 delta = 0.001 / kp->getMeanMotion() *  kp->getAltitude() / 3000.0;
      else if ( fabs( kp->getElevation() ) > 15.0 ) 
	 delta = 0.002 / kp->getMeanMotion() *  kp->getAltitude() / 3000.0;
      daynum += delta;

      printf("AAA %f %f %f %f %d\n", kp->getMeanMotion(), kp->getElevation(), daynum, delta, count );
   }
   else
   {
      if ( kp->getElevation() > 0.0 )
      {
	 if ( kp->getElevation() < 0.5 ) 
	    delta = 0.000167 / kp->getMeanMotion();
	 else if ( kp->getElevation() > 0.5 && kp->getElevation() < 5.0 ) 
	    delta = 0.002 / kp->getMeanMotion();
	 else if ( kp->getElevation() > 1.0 && kp->getElevation() < 10.0 ) 
	    delta = 0.01 / kp->getMeanMotion();
	 else if ( kp->getElevation() > 10.0 && kp->getElevation() < 25.0 ) 
	    delta = 0.02 / kp->getMeanMotion();
	 else if ( kp->getElevation() > 25.0 ) 
	    delta = 0.04 / kp->getMeanMotion();      
      }
      else
      {
	 if ( kp->getElevation() > -0.5 ) 
	    delta = 0.000167 / kp->getMeanMotion();
	 else if ( kp->getElevation() < -0.5 && kp->getElevation() > -5.0 ) 
	    delta = 0.002 / kp->getMeanMotion();
	 else if ( kp->getElevation() < -1.0 && kp->getElevation() > -10.0 ) 
	    delta = 0.02 / kp->getMeanMotion();
	 else if ( kp->getElevation() < -10.0 && kp->getElevation() > -25.0 ) 
	    delta = 0.04 / kp->getMeanMotion();
	 else if ( kp->getElevation() < -25.0 ) 
	    delta = 0.08 / kp->getMeanMotion();      	 
      }
      daynum += delta;
      //printf("BBB %f %f %f %f %d\n", kp->getMeanMotion(), kp->getElevation(), daynum, delta, count );
   }

   count++;

   return daynum;
}

void PetitTrack::computeEphem()
{
   char name[81], line1[81], line2[81], temp[256];

   // If we are already at work, don't bother.
   if ( calculating ) return;
   calculating = true;
   ephemStopFlag = false;
   
   strncpy( name, p[rt1ObjectID]->getSatName(), 80 );
   strncpy( line1, p[rt1ObjectID]->getTLELine1(), 80 );
   strncpy( line2, p[rt1ObjectID]->getTLELine2(), 80 );

   ephem = new KepPropagator();
   ephem->init( name, line1, line2, lat, lon, alt );

   ephemDaynum = getDaynum();

   if ( ephem->Geostationary() || !ephem->AosHappens() )
   {
      ephemView->clear();
      sprintf( temp, "There will be no passes for\n%s", ephem->getSatName() );
      ephemView->append( temp );
      sprintf( temp, "at your geographical coordinates.\n" );
      ephemView->append( temp );
      calculating = false;
      return;
   }
   
   ephemView->clear();
   sprintf( temp, "Ephemeris for %s\n", ephem->getSatName() );
   ephemView->append( temp );
   sprintf( temp, "From %s\n", ephem->daynum2String( ephemDaynum ) );
   ephemView->append( temp );
   sprintf( temp, "---- AOS ---- ---- LOS ---- EL  AZ" );
   ephemView->append( temp );

   ephemaos = 0;
   ephemlos = 0;
   ephemstart = ephemDaynum;
   ephemday = 0.0;
   epheminview = false;

   ephemTimer->start( 0, false );
}

void PetitTrack::runEphem()
{
   char aos_buf[40], los_buf[40], temp[80];
   time_t aos_t, los_t;
   struct tm *aos_tm, *los_tm;
   static int count = 0;
   static int maxEl = 0;
   static int maxElAz = 0;
   
   while ( count++ < 10 )
   {
      ephem->propagate( ephemDaynum );
      ephemday = ephemDaynum - ephemstart;

      if ( ephem->getElevation() > maxEl ) 
      {
	 maxEl = (int)ephem->getElevation();
	 maxElAz = (int)ephem->getAzimuth();
      }
      
      if ( !epheminview && ephem->getElevation() >= 0.0 )
      {
	 epheminview = true;
	 ephemaos = ephemDaynum;
      }
      
      if ( epheminview && ephem->getElevation() < 0.0 )
      {
	 epheminview = false;
	 ephemlos = ephemDaynum;
	 
	 aos_t = ephem->daynum2unix( ephemaos );
	 aos_tm = gmtime( &aos_t );
	 sprintf( aos_buf, "%02d%02d %02d:%02d:%02d",
		  aos_tm->tm_mon + 1, aos_tm->tm_mday, 
		  aos_tm->tm_hour, aos_tm->tm_min, aos_tm->tm_sec );
	 
	 los_t = ephem->daynum2unix( ephemlos );
	 los_tm = gmtime( &los_t );
	 sprintf( los_buf, "%02d%02d %02d:%02d:%02d",
		  los_tm->tm_mon + 1, los_tm->tm_mday, 
		  los_tm->tm_hour, los_tm->tm_min, los_tm->tm_sec );	 
	 sprintf( temp, "%s %s %2d %3d", aos_buf, los_buf, maxEl, maxElAz );
	 ephemView->append( temp );
	 ephemDaynum += 55.5E-3;
	 ephem->propagate( ephemDaynum );
	 maxEl = 0;
      }

      ephemDaynum = incrementDaynum( ephemDaynum, ephem );
      if ((int)ephemday > ephemDays || ephemStopFlag )
      {
	 ephemTimer->stop();
	 sprintf( temp, "-------------- END ---------------");
	 ephemView->append( temp );
	 calculating = false;
	 ephemStopFlag = false;
	 maxEl = 0;
	 delete ephem;
	 break;
      }
   }
   count = 0;
}

void PetitTrack::itemSelected( QListViewItem *listItem )
{
   char temp[30];
   int index;
   
   sprintf( temp, "%s", listItem->text( 4 ).latin1() );
   index = atoi( temp );
   
   setRT1Object( index );

   p[rt1ObjectID]->propagate( getDaynum() );   
   calculateGroundTrack();
   drawMap();   
   setTxRxDialogObject( rt1ObjectID );   
}

void PetitTrack::setTxRxDialogObject( int index )
{
   char temp[30];
   
   objTxRxIndex = index; 
   objTxRxLabel->setText( p[objTxRxIndex]->getSatName() );
   sprintf( temp, "%f", p[objTxRxIndex]->getUplink() );
   txfLine->setText( temp );
   sprintf( temp, "%f", p[objTxRxIndex]->getDownlink() );
   rxfLine->setText( temp );      
}

void PetitTrack::calculateGroundTrack()
{
   KepPropagator *gtp;
   char name[81], line1[81], line2[81];
   struct timeb tptr;
   int t;
   double groundTrackDaynum;
   int x, y;
   double azimuth;
   
   strncpy( name, p[rt1ObjectID]->getSatName(), 80 );
   strncpy( line1, p[rt1ObjectID]->getTLELine1(), 80 );
   strncpy( line2, p[rt1ObjectID]->getTLELine2(), 80 );
   
   gtp = new KepPropagator();
   gtp->init( name, line1, line2, lat, lon, alt );
   
   ftime( &tptr );
   t = tptr.time / 120;
   t = t * 120;
   
   groundTrackDaynum = ((((double)t + 0.001 * (double)tptr.millitm) / 86400.0 ) - 3651.0);
   groundTrackDaynum -= 41.44447E-3; //Subtract 60 minutes
   
   for ( int i = 0; i < 60; i++ )
   {
      gtp->propagate( groundTrackDaynum );
      groundTrackDaynum += 1.38889E-3; // Increment 2 minutes
      if ( (int)gtp->getSspLon() <= 180 ) 
	 x = (int)((float)worldPix->height() - (float)worldPix->width() * gtp->getSspLon() / 360.0 );
      else 
	 x = (int)((float)worldPix->width() - (float)worldPix->width() * ( gtp->getSspLon() - 180.0 ) / 360.0 );
      
      y = (int)( (float)worldPix->height() / 2.0 - (float)worldPix->height() / 2.0 * gtp->getSspLat() / 90.0 );
      groundTrackPoints->setPoint( i, x, y );
      
      // Save the az and el for the sky view
      if ( gtp->getElevation() > 0 )
      {
	 azimuth=(double)( gtp->getAzimuth() - 90.0) * deg2rad;
	 x=40 + (int)((149/2.0) - cos(azimuth)*((double)(149/2.0)-gtp->getElevation()*(double)(149/2.0)/90.0));
	 y=15 + (int)((149/2.0) + sin(azimuth)*((double)(149/2.0)-gtp->getElevation()*(double)(149/2.0)/90.0));
	 AzElPoints->setPoint( i, x, y );
      }
      else  AzElPoints->setPoint( i, -1, -1 );
   }
   groundTrackFlag = true;
   delete gtp;
}

void PetitTrack::calculateSunFootprint()
{
   int azi, xx, yy;
   float ssplat, ssplong, azimuth;
   float num, dem, beta, spiderlat, spiderlong;
   float height = (float)worldPix->height();
   float width = (float)worldPix->width();
   float TWOPI = 2.0 * M_PI;
   float deg2rad = 1.74532925199e-02, R0=6378.16;
   double daynum;
   
   daynum = getDaynum();
   
   sun->propagate( daynum );
   moon->propagate( daynum );
   ssplat = sun->getLat() * deg2rad;
   ssplong = sun->getLon() * deg2rad;
   beta = 0.5 * 20000 / R0;

   /* Compute the visibility circle */
   for ( azi = 0; azi < 360; azi++)
   {
      azimuth = deg2rad * (double)azi;
      spiderlat = asin( sin( ssplat ) * cos( beta )
			+ cos( azimuth ) * sin( beta ) * cos( ssplat ));
      num = cos( beta ) - ( sin( ssplat ) * sin( spiderlat ));
      dem = cos( ssplat ) * cos( spiderlat );

      if ( azi == 0 && ( beta > M_PI_2 - ssplat ))
	 spiderlong = ssplong + M_PI;
      else if ( azi == 180 && ( beta > M_PI_2 + ssplat ))
	 spiderlong = ssplong + M_PI;
      else if ( fabs( num / dem) > 1.0 )
	 spiderlong = ssplong;
      else
      {
	 if (( 180 - azi ) >= 0)
	    spiderlong = ssplong - arccos( num, dem );
	 else
	    spiderlong = ssplong + arccos( num, dem );
      }

      while ( spiderlong < 0.0 )
	 spiderlong += TWOPI;

      while ( spiderlong > ( 2.0 * M_PI ))
	 spiderlong -= TWOPI;

      spiderlat = spiderlat / deg2rad;
      spiderlong = spiderlong / deg2rad;

      /* Convert spiderweb data to map-based coordinates */
      if ( spiderlong <= 180.0 )
	 xx = (int)( height - ( width * ( spiderlong / 360.0 )));
      else
      {
	 spiderlong = spiderlong - 180.0;
	 xx = (int)( width - ( width * ( spiderlong / 360.0 )));
      }

      yy = (int)( height / 2.0 - ( height / 2.0 * ( spiderlat / 90.0 )));
      sunPoints->setPoint( azi, xx, yy );
   }   

   // The sun map coordinates 
   if ( sun->getLon() <= 180.0) sun_x = (int)( height - (width * ( sun->getLon() / 360.0 )));
   else sun_x = (int)( width - ( width * ( (sun->getLon() - 180.0) / 360.0 )));

   sun_y = (int)( height / 2.0 - ( height / 2.0 * ( sun->getLat() / 90.0 )));

   // The moon map corrdinates
   if ( moon->getLon() <= 180.0) moon_x = (int)( height - (width * ( moon->getLon() / 360.0 )));
   else moon_x = (int)( width - ( width * ( (moon->getLon() - 180.0) / 360.0 )));

   moon_y = (int)( height / 2.0 - ( height / 2.0 * ( moon->getLat() / 90.0 )));

}

void PetitTrack::setEphemDays( int days )
{
   ephemDays = days;
}

void PetitTrack::stopEphem()
{
   ephemStopFlag = true;
}

void PetitTrack::setMapType( int type )
{
   switch( type )
   {
   case 0:
      mapType = 0;
      worldPix = map0;
      break;
   case 1:
      mapType = 1;
      worldPix = map1;
      break;
   case 2:
      mapType = 2;
      worldPix = map2;
      break;
   }
   
   drawMap();
}

void PetitTrack::objTxRxNavigateLeft() 
{
   char temp[20];
   bool ok;
   
   txfLine->text().toDouble( &ok );
   if ( ok ) p[objTxRxIndex]->setUplink( atof( txfLine->text() ) );

   rxfLine->text().toDouble( &ok );
   if ( ok ) p[objTxRxIndex]->setDownlink( atof( rxfLine->text() ) );
					  					   
   if ( --objTxRxIndex < 0 )
      objTxRxIndex = numObjects - 1;
   
   objTxRxLabel->setText( p[objTxRxIndex]->getSatName() );
   sprintf( temp, "%f", p[objTxRxIndex]->getUplink() );
   txfLine->setText( temp );
   sprintf( temp, "%f", p[objTxRxIndex]->getDownlink() );
   rxfLine->setText( temp );
}

void PetitTrack::objTxRxNavigateRight() 
{
   char temp[20];
   bool ok;

   txfLine->text().toDouble( &ok );
   if ( ok ) p[objTxRxIndex]->setUplink( atof( txfLine->text() ) );

   rxfLine->text().toDouble( &ok );
   if ( ok ) p[objTxRxIndex]->setDownlink( atof( rxfLine->text() ) );
					  					   
   if ( ++objTxRxIndex >= numObjects )
      objTxRxIndex = 0;
   
   objTxRxLabel->setText( p[objTxRxIndex]->getSatName() );   
   sprintf( temp, "%f", p[objTxRxIndex]->getUplink() );
   txfLine->setText( temp );
   sprintf( temp, "%f", p[objTxRxIndex]->getDownlink() );
   rxfLine->setText( temp );
}

void PetitTrack::tabChanged( QWidget * )
{
   static int prevIndex = 0;
   bool ok;
   
   if ( tab->currentPageIndex() != 4 && prevIndex == 4 )
   {
      txfLine->text().toDouble( &ok );
      if ( ok ) p[objTxRxIndex]->setUplink( atof( txfLine->text() ) );
      
      rxfLine->text().toDouble( &ok );
      if ( ok ) p[objTxRxIndex]->setDownlink( atof( rxfLine->text() ) );
   }
   prevIndex = tab->currentPageIndex();
}

